@@ -22,6 +22,10 @@ try:
     intern = intern
 except NameError:
     intern = sys.intern
+try:
+    basestring = basestring
+except NameError:
+    basestring = str
 
 supports_symlinks = True
 try:
@@ -45,6 +49,14 @@ __all__ = [
 # Internals
 #
 
+_py2 = sys.version_info < (3,)
+_py2_fs_encoding = 'ascii'
+
+def _py2_fsencode(parts):
+    # py2 => minimal unicode support
+    return [part.encode(_py2_fs_encoding) if isinstance(part, unicode)
+            else part for part in parts]
+
 def _is_wildcard_pattern(pat):
     # Whether this pattern needs actual matching using fnmatch, or can
     # be looked up directly as a file.
@@ -59,6 +71,8 @@ class _Flavour(object):
         self.join = self.sep.join
 
     def parse_parts(self, parts):
+        if _py2:
+            parts = _py2_fsencode(parts)
         parsed = []
         sep = self.sep
         altsep = self.altsep
@@ -263,42 +277,47 @@ class _PosixFlavour(_Flavour):
 
     def resolve(self, path):
         sep = self.sep
-        def split(p):
-            return [x for x in p.split(sep) if x]
-        def absparts(p):
-            # Our own abspath(), since the posixpath one makes
-            # the mistake of "normalizing" the path without resolving the
-            # symlinks first.
-            if not p.startswith(sep):
-                return split(os.getcwd()) + split(p)
-            else:
-                return split(p)
-        parts = absparts(str(path))[::-1]
         accessor = path._accessor
-        resolved = cur = ""
-        symlinks = {}
-        while parts:
-            part = parts.pop()
-            cur = resolved + sep + part
-            if cur in symlinks and symlinks[cur] <= len(parts):
-                # We've already seen the symlink and there's not less
-                # work to do than the last time.
-                raise RuntimeError("Symlink loop from %r" % cur)
-            try:
-                target = accessor.readlink(cur)
-            except OSError as e:
-                if e.errno != EINVAL:
-                    raise
-                # Not a symlink
-                resolved = cur
-            else:
-                # Take note of remaining work from this symlink
-                symlinks[cur] = len(parts)
-                if target.startswith(sep):
-                    # Symlink points to absolute path
-                    resolved = ""
-                parts.extend(split(target)[::-1])
-        return resolved or sep
+        seen = {}
+        def _resolve(path, rest):
+            if rest.startswith(sep):
+                path = ''
+
+            for name in rest.split(sep):
+                if not name or name == '.':
+                    # current dir
+                    continue
+                if name == '..':
+                    # parent dir
+                    path, _, _ = path.rpartition(sep)
+                    continue
+                newpath = path + sep + name
+                if newpath in seen:
+                    # Already seen this path
+                    path = seen[newpath]
+                    if path is not None:
+                        # use cached value
+                        continue
+                    # The symlink is not resolved, so we must have a symlink loop.
+                    raise RuntimeError("Symlink loop from %r" % newpath)
+                # Resolve the symbolic link
+                try:
+                    target = accessor.readlink(newpath)
+                except OSError as e:
+                    if e.errno != EINVAL:
+                        raise
+                    # Not a symlink
+                    path = newpath
+                else:
+                    seen[newpath] = None # not resolved symlink
+                    path = _resolve(path, target)
+                    seen[newpath] = path # resolved symlink
+
+            return path
+        # NOTE: according to POSIX, getcwd() cannot contain path components
+        # which are symlinks.
+        base = '' if path.is_absolute() else os.getcwd()
+        return _resolve(base, str(path)) or sep
 
     def is_reserved(self, parts):
         return False
@@ -578,8 +597,7 @@ class PurePath(object):
         for a in args:
             if isinstance(a, PurePath):
                 parts += a._parts
-            elif isinstance(a, str):
-                # Assuming a str
+            elif isinstance(a, basestring):
                 parts.append(a)
             else:
                 raise TypeError(
@@ -762,6 +780,12 @@ class PurePath(object):
     def with_suffix(self, suffix):
         """Return a new path with the file suffix changed (or added, if none)."""
         # XXX if suffix is None, should the current suffix be removed?
+        drv, root, parts = self._flavour.parse_parts((suffix,))
+        if drv or root or len(parts) != 1:
+            raise ValueError("Invalid suffix %r" % (suffix))
+        suffix = parts[0]
+        if not suffix.startswith('.'):
+            raise ValueError("Invalid suffix %r" % (suffix))
         name = self.name
         if not name:
             raise ValueError("%r has an empty name" % (self,))
@@ -787,27 +811,23 @@ class PurePath(object):
         parts = self._parts
         drv = self._drv
         root = self._root
-        if drv or root:
-            if root:
-                abs_parts = [drv, root] + parts[1:]
-            else:
-                abs_parts = [drv] + parts[1:]
+        if root:
+            abs_parts = [drv, root] + parts[1:]
         else:
             abs_parts = parts
         to_drv, to_root, to_parts = self._parse_args(other)
-        if to_drv or to_root:
-            if to_root:
-                to_abs_parts = [to_drv, to_root] + to_parts[1:]
-            else:
-                to_abs_parts = [to_drv] + to_parts[1:]
+        if to_root:
+            to_abs_parts = [to_drv, to_root] + to_parts[1:]
         else:
             to_abs_parts = to_parts
         n = len(to_abs_parts)
-        if n == 0 and (drv or root) or abs_parts[:n] != to_abs_parts:
+        cf = self._flavour.casefold_parts
+        if (root or drv) if n == 0 else cf(abs_parts[:n]) != cf(to_abs_parts):
             formatted = self._format_parsed_parts(to_drv, to_root, to_parts)
             raise ValueError("{!r} does not start with {!r}"
                              .format(str(self), str(formatted)))
-        return self._from_parsed_parts('', '', abs_parts[n:])
+        return self._from_parsed_parts('', root if n == 1 else '',
+                                       abs_parts[n:])
 
     @property
     def parts(self):
@@ -838,7 +858,6 @@ class PurePath(object):
     if sys.version_info < (3,):
         __div__ = __truediv__
         __rdiv__ = __rtruediv__
-        del __truediv__, __rtruediv__
 
     @property
     def parent(self):
@@ -1088,7 +1107,7 @@ class Path(PurePath):
             except OSError as e:
                 if e.errno != ENOENT:
                     raise
-                self.parent.mkdir(mode, True)
+                self.parent.mkdir(parents=True)
                 self._accessor.mkdir(self, mode)
 
     def chmod(self, mode):
